from decimal import Decimal
from typing import Optional, List

import edclasses.api_adapters.elite_bgs_adapter as bgs_adapter
from . import enums
from .utils import UniqueInstanceMixin, OneToManyRelation, AutoRefreshMixin


EXPIRATION_TIME_MINUTES = 20


class System(UniqueInstanceMixin, AutoRefreshMixin):
    keys = ("name",)
    registry = {}
    adapter = bgs_adapter.EliteBgsSystemAdapter()
    refreshed_fields = ("faction_branches", "stations", "eddb_id",)
    EXPIRATION_TIME_MINUTES = EXPIRATION_TIME_MINUTES

    _stations_relation = OneToManyRelation.create(
        parent_class_name="System", child_class_name="OrbitalStation"
    )
    _faction_branches_relation = OneToManyRelation.create(
        parent_class_name="System", child_class_name="FactionBranch"
    )

    def __init__(self, name: str, stations=None, faction_branches=None, eddb_id=None):
        self.eddb_id = eddb_id
        self.name = name
        self.stations = stations or []
        self.faction_branches = faction_branches or []
        super().__init__()

    def __repr__(self):
        return f"System '{self.name}'"

    def _stations_setter(self, value):
        self._stations_relation.set_for_parent(self, value)

    def _stations_getter(self):
        return self._stations_relation.get_for_parent(self)

    stations = property(fget=_stations_getter, fset=_stations_setter)

    def _faction_branches_setter(self, value):
        self._faction_branches_relation.set_for_parent(self, value)

    def _faction_branches_getter(self):
        return self._faction_branches_relation.get_for_parent(self)

    faction_branches = property(
        fget=_faction_branches_getter, fset=_faction_branches_setter
    )


class Faction(UniqueInstanceMixin, AutoRefreshMixin):
    keys = ("name",)
    registry = {}
    adapter = bgs_adapter.EliteBgsFactionAdapter()
    refreshed_fields = ("faction_branches",)
    EXPIRATION_TIME_MINUTES = EXPIRATION_TIME_MINUTES

    _faction_branches_relation = OneToManyRelation.create(
        parent_class_name="Faction", child_class_name="FactionBranch"
    )

    def __init__(self, name: str):
        self.name = name
        super().__init__()

    def __repr__(self):
        return f"Faction '{self.name}'"

    def _faction_branches_setter(self, value):
        self._faction_branches_relation.set_for_parent(self, value)

    def _faction_branches_getter(self):
        return self._faction_branches_relation.get_for_parent(self)

    faction_branches = property(
        fget=_faction_branches_getter, fset=_faction_branches_setter
    )


class FactionBranch(UniqueInstanceMixin, AutoRefreshMixin):
    keys = (
        "faction",
        "system",
    )
    registry = {}
    adapter = bgs_adapter.EliteBgsFactionBranchAdapter()
    refreshed_fields = ("influence", "stations")
    EXPIRATION_TIME_MINUTES = EXPIRATION_TIME_MINUTES

    _system_relation = OneToManyRelation.create(
        parent_class_name="System", child_class_name="FactionBranch"
    )
    _faction_relation = OneToManyRelation.create(
        parent_class_name="Faction", child_class_name="FactionBranch"
    )
    _stations_relation = OneToManyRelation.create(
        parent_class_name="FactionBranch", child_class_name="OrbitalStation"
    )

    def __init__(
            self,
            faction: Faction,
            system: System,
            is_main: bool = False,
            influence: Decimal = None,
            stations: List = None,
    ):
        self.faction = faction
        self.system = system
        self.is_main = is_main
        self.influence = influence
        self.stations = stations or []
        super().__init__()

    def __repr__(self):
        return f"{self.faction} in {self.system}"

    def _system_setter(self, value):
        self._system_relation.set_for_child(self, value)

    def _system_getter(self):
        return self._system_relation.get_for_child(self)

    system = property(fget=_system_getter, fset=_system_setter)

    def _faction_setter(self, value):
        self._faction_relation.set_for_child(self, value)

    def _faction_getter(self):
        return self._faction_relation.get_for_child(self)

    faction = property(fget=_faction_getter, fset=_faction_setter)

    def _stations_setter(self, value):
        self._stations_relation.set_for_parent(self, value)

    def _stations_getter(self):
        return self._stations_relation.get_for_parent(self)

    stations = property(fget=_stations_getter, fset=_stations_setter)


class OrbitalStation(UniqueInstanceMixin, AutoRefreshMixin):
    keys = (
        "name",
        "system",
    )
    registry = {}
    adapter = bgs_adapter.EliteBgsStationAdapter()
    refreshed_fields = ("controlling_faction", "distance_to_arrival", "services")
    EXPIRATION_TIME_MINUTES = EXPIRATION_TIME_MINUTES

    _system_relation = OneToManyRelation.create(
        parent_class_name="System", child_class_name="OrbitalStation"
    )
    _controlling_faction_relation = OneToManyRelation.create(
        parent_class_name="FactionBranch", child_class_name="OrbitalStation"
    )

    def __init__(
            self,
            name: str,
            station_type: enums.StationType,
            system: System,
            distance_to_arrival: Optional[Decimal] = None,
            services: Optional[List] = None,
            controlling_faction=None,
    ):
        self.name = name
        self.station_type = station_type.value
        self.system = system
        self.distance_to_arrival = distance_to_arrival
        self.services = services or []
        self.controlling_faction = None or controlling_faction
        super().__init__()

    def __repr__(self):
        return f"{self.station_type.title()} '{self.name}'"

    def _system_setter(self, value):
        self._system_relation.set_for_child(self, value)

    def _system_getter(self):
        return self._system_relation.get_for_child(self)

    system = property(fget=_system_getter, fset=_system_setter)

    def _controlling_faction_setter(self, value):
        self._controlling_faction_relation.set_for_child(self, value)

    def _controlling_faction_getter(self):
        return self._controlling_faction_relation.get_for_child(self)

    controlling_faction = property(
        fget=_controlling_faction_getter, fset=_controlling_faction_setter
    )



import factory

from ...classes import System, Faction, FactionBranch, OrbitalStation
from ...enums import StationType


class SystemFactory(factory.Factory):
    class Meta:
        model = System

    name = factory.Sequence(lambda n: f"System {n}")


class FactionFactory(factory.Factory):
    class Meta:
        model = Faction

    name = factory.Sequence(lambda n: f"Faction {n}")


class FactionBranchFactory(factory.Factory):
    class Meta:
        model = FactionBranch

    faction = factory.SubFactory(FactionFactory)
    system = factory.SubFactory(SystemFactory)
    stations = list()
    is_main = False  # we don't need to know faction's main branch - usually we won't have this information
    influence = 50


class OrbitalStationFactory(factory.Factory):
    class Meta:
        model = OrbitalStation

    name = factory.Sequence(lambda n: f"Orbital Station {n}")
    station_type = StationType.STATION
    system = factory.SubFactory(SystemFactory)
    distance_to_arrival = 100
    services = list()
    controlling_faction = factory.SubFactory(FactionBranchFactory)



from decimal import Decimal
from typing import List, Optional

from .utils import get_orbital_station, get_faction_branch, get_faction, get_system
from .. import enums
from ..api_clients import EliteBgsClient
from ..utils import return_first_match

ELITE_BGS_CLIENT = EliteBgsClient()


class EliteBgsAdapterBase:
    def __init__(self):
        self.client = ELITE_BGS_CLIENT

    def _convert_station_dict_to_obj(self, station_dict: dict) -> "OrbitalStation":
        station_type = station_dict["type"]
        try:
            station_type_enum = enums.StationType(station_type)
        except ValueError:
            station_type_enum = enums.StationType.STATION

        return get_orbital_station(
            name=station_dict["name"],
            station_type=station_type_enum,
            system=station_dict["system"],
            distance_to_arrival=station_dict["distance_from_star"],
        )

    def _convert_faction_presence_dict_to_obj(
        self, faction_presence_dict: dict, faction_name: str
    ) -> "FactionBranch":
        faction = get_faction(name=faction_name)
        system = get_system(name=faction_presence_dict["system_name"])
        faction_branch = get_faction_branch(faction=faction, system=system)
        return faction_branch

    @staticmethod
    def _get_factions_from_response(response: dict) -> List:
        return response.get("docs", [])


class EliteBgsFactionBranchAdapter(EliteBgsAdapterBase):
    def influence(self, faction_branch: "FactionBranch") -> Decimal:
        faction_name = faction_branch.faction.name
        data = self.client.factions(system=faction_branch.system.name)

        factions = data["docs"]
        faction = return_first_match(
            lambda fact: fact["name"].lower() == faction_name.lower(), factions
        )
        faction_presence_list = faction["faction_presence"]
        faction_presence = return_first_match(
            lambda fact: fact["system_name_lower"]
            == faction_branch.system.name.lower(),
            faction_presence_list,
        )

        return Decimal(faction_presence["influence"])

    def stations(self, faction_branch: "FactionBranch") -> List["OrbitalStation"]:
        faction_name = faction_branch.faction.name
        system_name = faction_branch.system.name
        data = self.client.stations(system=system_name)
        stations = data["docs"]

        faction_stations = filter(
            lambda station: station["controlling_minor_faction"].lower()
            == faction_name.lower(),
            stations,
        )

        station_objects = []
        for station in faction_stations:
            station_obj = self._convert_station_dict_to_obj(station)
            station_objects.append(station_obj)

        return station_objects


class EliteBgsSystemAdapter(EliteBgsAdapterBase):
    def faction_branches(self, system: "System"):
        system_name_lower = system.name.lower()
        # TODO: to save requests, it would be better to use client.system with factionDetails=True instead of factions,
        # because it's also used in system.eddb_id
        data = self.client.factions(system=system.name)
        factions = data["docs"]

        faction_branches = []
        for faction in factions:
            faction_in_this_system = return_first_match(
                lambda fact: fact["system_name_lower"] == system_name_lower,
                faction["faction_presence"],
            )
            faction_branch_obj = self._convert_faction_presence_dict_to_obj(
                faction_presence_dict=faction_in_this_system,
                faction_name=faction["name"],
            )
            faction_branches.append(faction_branch_obj)

        return faction_branches

    def stations(self, system: "System") -> List["OrbitalStations"]:
        data = self.client.stations(system=system.name)
        stations = data["docs"]

        station_objects = []
        for station in stations:
            station_obj = self._convert_station_dict_to_obj(station)
            station_objects.append(station_obj)

        return station_objects

    def eddb_id(self, system: "System") -> Optional[int]:
        # TODO: this could be taken from self.client.factions or stations - this way we would get more data.
        data = self.client.systems(name=system.name)
        systems = data["docs"]
        system_data = return_first_match(
            lambda syst: syst["name_lower"] == system.name.lower(),
            systems,
        )
        if system_data:
            return system_data["eddb_id"]
        return None



class EliteBgsFactionAdapter(EliteBgsAdapterBase):
    def faction_branches(self, faction_obj):
        data = self.client.factions(name=faction_obj.name)
        factions = data["docs"]

        faction_branches = []

        faction_data = return_first_match(
            lambda fact: fact["name"].lower() == faction_obj.name.lower(), factions
        )

        for faction_presence in faction_data["faction_presence"]:
            faction_branch_obj = self._convert_faction_presence_dict_to_obj(
                faction_presence_dict=faction_presence,
                faction_name=faction_obj.name,
            )
            faction_branches.append(faction_branch_obj)

        return faction_branches


class EliteBgsStationAdapter(EliteBgsAdapterBase):
    def distance_to_arrival(self, obj):
        distance_to_arrival = self._get_this_station_data(
            obj=obj, key="distance_from_star"
        )

        if distance_to_arrival is not None:
            return Decimal(distance_to_arrival)
        return None

    def services(self, obj):
        services_data = self._get_this_station_data(obj=obj, key="services")

        services = [
            enums.StationService(service_dict["name_lower"])
            for service_dict in services_data
        ]
        return services

    def controlling_faction(self, obj):
        controlling_faction_name = self._get_this_station_data(
            obj=obj,
            key="controlling_minor_faction_cased"
        )
        faction = get_faction(controlling_faction_name)
        faction_branch = get_faction_branch(faction, obj.system)

        return faction_branch

    def _get_this_station_data(self, obj: "OrbitalStation", key: str):
        data = self.client.stations(system=obj.system.name)

        stations = data["docs"]
        this_station_data = return_first_match(
            lambda station: station["name"].lower() == obj.name.lower(), stations
        )
        return this_station_data[key]



from decimal import Decimal

import edclasses


def get_orbital_station(
    name,
    station_type,
    system,
    distance_to_arrival=None,
    services=None,
    controlling_faction=None,
):

    if isinstance(system, str):
        system = edclasses.System.create(name=system)

    if distance_to_arrival and not isinstance(distance_to_arrival, Decimal):
        distance_to_arrival = Decimal(distance_to_arrival)

    return edclasses.OrbitalStation.create(
        name=name,
        station_type=station_type,
        system=system,
        distance_to_arrival=distance_to_arrival,
        services=services,
        controlling_faction=controlling_faction,
    )


def get_faction(name):
    return edclasses.Faction.create(name=name)


def get_faction_branch(faction, system):
    if isinstance(system, str):
        system = edclasses.System.create(name=system)

    if isinstance(faction, str):
        faction = edclasses.Faction.create(name=faction)

    return edclasses.FactionBranch.create(faction=faction, system=system)


def get_system(name):
    return edclasses.System.create(name=name)



from enum import Enum


class StationType(Enum):
    # these are types as found in Elite Bgs API.
    CRATER_OUTPOST = "crateroutpost"
    ORBIS = "orbis"
    CORIOLIS = "coriolis"
    BERNAL = "bernal"
    ASTEROID_BASE = "asteroidbase"
    OUTPOST = "outpost"
    CRATER_PORT = "craterport"
    OCELLUS = "ocellus"
    ON_FOOT_SETTLEMENT = "onfootsettlement"
    STATION = "station"  # fallback for unknown type
    # MEGASHIP = "mega ship" # TODO: handle megaships


class StationService(Enum):
    PIONEERSUPPLIES = "pioneersupplies"
    SHOP = "shop"
    CONTACTS = "contacts"
    FLIGHTCONTROLLER = "flightcontroller"
    FACILITATOR = "facilitator"
    MATERIALTRADER = "materialtrader"
    APEXINTERSTELLAR = "apexinterstellar"
    CARRIERVENDOR = "carriervendor"
    AUTODOCK = "autodock"
    MISSIONSGENERATED = "missionsgenerated"
    STATIONOPERATIONS = "stationoperations"
    TUNING = "tuning"
    CREWLOUNGE = "crewlounge"
    POWERPLAY = "powerplay"
    ENGINEER = "engineer"
    SHIPYARD = "shipyard"
    SEARCHRESCUE = "searchrescue"
    REPAIR = "repair"
    COMMODITIES = "commodities"
    STATIONMENU = "stationmenu"
    MODULEPACKS = "modulepacks"
    TECHBROKER = "techbroker"
    SOCIALSPACE = "socialspace"
    DOCK = "dock"
    FRONTLINESOLUTIONS = "frontlinesolutions"
    LIVERY = "livery"
    OUTFITTING = "outfitting"
    REFUEL = "refuel"
    EXPLORATION = "exploration"
    BLACKMARKET = "blackmarket"
    MISSIONS = "missions"
    REARM = "rearm"
    BARTENDER = "bartender"
    VISTAGENOMICS = "vistagenomics"



import datetime
from typing import Optional, Any

_NEVER_EXPIRE = 0  # TODO: this could be obj()
NOT_SET = object()


class ExpiringCachedPropertyMixin:
    """
    To be used with @cached_property_decorator.

    Add this mixin to the class, if you want its cached property to expire after a set time.

    To mark a method as expiring property:

    1. Add @functools.cached_property decorator on the method which you want to cache for a limited period of time.
    2. Add attribuge 'expiring_properties_registry={}' to your class.
    3. Register your method in the expiring_properties_registry as {'<method_name>': <lifetime_in_seconds>}.

    When you first call obj.<name_of_cached_property>, it will be calculated and saved with an expiration date.
    When you call obj.<name_of_cached_property> and the time has expired, it will be recalculated.
    When you set your property manually, or pass it during class initialization, it will never expire.
    If you set None as lifetime_in_seconds, it will never expire.
    You can use self.clear_property(property_name) to manually clear the cache on given attribute
    """

    def _clear_property(self, item: str) -> None:
        get_attr = super().__getattribute__
        try:
            registry = get_attr("expiring_properties_registry")
        except AttributeError:
            raise NotImplemented(
                f"You must define expiring_properties_registry on {self.__class__.__name__} to use"
                " ExpiringCachedPropertyMixIn"
            )
        if item not in registry:
            raise ValueError(f"{self.__class__.__name__}.{item} is registered!")

        cache = get_attr("__dict__")
        cache.pop(item, None)
        expire_key = self._get_expiration_key(item)
        cache.pop(expire_key, None)

    @staticmethod
    def _get_expiration_key(item: str) -> str:
        return f"{item}_expiration_time"

    def _get_new_expiration_time(
        self, lifetime_in_seconds: int
    ) -> Optional[datetime.datetime]:
        if lifetime_in_seconds:
            return datetime.datetime.utcnow() + datetime.timedelta(
                seconds=lifetime_in_seconds  # TODO: change to minutes
            )
        else:
            return None

    @staticmethod
    def _is_expired(expiration_time: [datetime.datetime, int]) -> bool:
        if expiration_time and datetime.datetime.utcnow() >= expiration_time:
            return True
        return False

    def __getattribute__(self, item: str) -> Any:
        get_attr = super().__getattribute__
        try:
            registry = get_attr("expiring_properties_registry")
        except AttributeError:
            raise NotImplemented(
                f"You must define expiring_properties_registry on {self.__class__.__name__} to use"
                " ExpiringCachedPropertyMixIn"
            )

        try:
            item_lifetime = registry[item]
        except KeyError:
            return get_attr(item)

        cache = get_attr("__dict__")
        expiration_key = get_attr("_get_expiration_key")(item)
        expiration_time = cache.get(expiration_key)
        time_expired = expiration_time and datetime.datetime.utcnow() >= expiration_time
        if time_expired or cache[item] == NOT_SET:
            cache.pop(item, None)

        val = get_attr(item)

        if expiration_time is None or time_expired:
            cache[expiration_key] = get_attr("_get_new_expiration_time")(
                lifetime_in_seconds=item_lifetime
            )
        return val

    def __setattr__(self, key: str, value: Any) -> None:
        get_attr = super().__getattribute__

        try:
            registry = get_attr("expiring_properties_registry")
        except AttributeError:
            raise NotImplemented(
                f"You must define expiring_properties_registry on {self.__class__.__name__} to use"
                " ExpiringCachedPropertyMixIn"
            )

        __dict__ = get_attr("__dict__")
        if key in registry:
            expiration_key = get_attr("_get_expiration_key")(key)
            __dict__[expiration_key] = _NEVER_EXPIRE

        return super().__setattr__(key, value)



from urllib import parse


import requests
from functools import lru_cache

from ratelimit import limits


class EliteBgsClient:
    API_URL = "https://elitebgs.app/api/ebgs/v5/"

    def __init__(self):
        self.session = requests.Session()

    @limits(calls=20, period=60)
    def _get_request(self, path="", **kwargs):
        url = parse.urljoin(self.API_URL, path)
        response = self.session.get(url, params=kwargs)
        return response.json()

    @lru_cache  # TODO: replace this with a proper cache:
    # we should have a time expiring cache, would be best if we would save single objects from API response
    # (e.g. object representing faction presence) and then reuse that.
    def get_request(self, path="", **kwargs):
        return self._get_request(path, **kwargs)

    def factions(self, **kwargs):
        return self.get_request("factions", **kwargs)

    def stations(self, **kwargs):
        return self.get_request("stations", **kwargs)

    def systems(self, **kwargs):
        return self.get_request("systems", **kwargs)

    def ticks(self, **kwargs):
        return self.get_request("ticks", **kwargs)



import datetime
from collections import defaultdict
from typing import Tuple, List


def return_first_match(func, items):
    return next(item for item in items if func(item))


class UniqueInstanceMixin:
    registry = {}
    keys = tuple()

    @classmethod
    def create(cls, **kwargs):
        try:
            return cls(**kwargs)
        except InstanceAlreadyExists:
            return cls.get_from_registry(**kwargs)

    @classmethod
    def _get_key(cls, *args, **kwargs):
        if args:
            instance = args[0]
            return tuple(getattr(instance, attr) for attr in cls.keys)
        else:
            return tuple(kwargs[attr] for attr in cls.keys)

    @classmethod
    def get_from_registry(cls, **kwargs):
        obj_key = cls._get_key(**kwargs)
        return cls.registry.get(obj_key)

    def __init__(self, *args, **kwargs):
        obj_key = self._get_key(self)
        obj = self.__class__.registry.get(obj_key)
        if obj is not None:
            raise InstanceAlreadyExists
        self.__class__.registry[obj_key] = self
        super().__init__()

    def remove_from_registry(self):
        obj_key = self._get_key(**self.__dict__)
        self.registry.pop(obj_key)

    def delete(self):
        self.remove_from_registry()


class OneToOneRelation(UniqueInstanceMixin):
    registry = {}
    keys = (
        "parent_class_name",
        "child_class_name",
    )

    def __init__(self, parent_class_name: str, child_class_name: str):
        self.parent_class_name = parent_class_name
        self.child_class_name = child_class_name
        self.parent_side = {}
        self.child_side = {}
        super().__init__()

    def set_for_parent(self, parent_obj, child_obj):
        old_child = self.get_for_parent(parent_obj)

        if old_child and child_obj is not old_child:
            self._delete_link(parent_obj, old_child)

        if child_obj is not None:
            self._add_link(parent_obj, child_obj)

    def get_for_parent(self, parent_obj):
        return self.parent_side.get(parent_obj, None)

    def set_for_child(self, child_obj, parent_obj):
        old_parent = self.get_for_child(child_obj)

        if old_parent and parent_obj is not old_parent:
            self._delete_link(old_parent, child_obj)

        if parent_obj is not None:
            self._add_link(parent_obj, child_obj)

    def get_for_child(self, child_obj):
        return self.child_side.get(child_obj, None)

    def _delete_link(self, parent_obj, child_obj):
        self.parent_side.pop(parent_obj)
        self.child_side.pop(child_obj)

    def _add_link(self, parent_obj, child_obj):
        self.parent_side[parent_obj] = child_obj
        self.child_side[child_obj] = parent_obj


class OneToManyRelation(OneToOneRelation):
    registry = {}

    def __init__(self, parent_class_name: str, child_class_name: str):
        super().__init__(parent_class_name, child_class_name)
        self.parent_side = defaultdict(list)

    def _add_link(self, parent_obj, child_obj):
        self.parent_side[parent_obj].append(child_obj)
        self.child_side[child_obj] = parent_obj

    def _delete_link(self, parent_obj, child_obj):
        self.parent_side[parent_obj].remove(child_obj)
        self.child_side.pop(child_obj)

    def get_for_parent(self, parent_obj):
        return self.parent_side.get(parent_obj, list())

    def set_for_parent(self, parent_obj, children: List):
        old_children = self.get_for_parent(parent_obj)
        for child in old_children:
            self.child_side.pop(child)

        for child in children:
            old_parent = self.get_for_child(child)
            if old_parent is not None:
                self._delete_link(old_parent, child)
            self.child_side[child] = parent_obj

        self.parent_side[parent_obj] = children

    def set_for_child(self, child_obj, parent_obj):
        old_parent = self.get_for_child(child_obj)

        if parent_obj is old_parent:
            return

        if old_parent is not None:
            self._delete_link(old_parent, child_obj)

        if parent_obj is not None:
            self._add_link(parent_obj, child_obj)


class InstanceAlreadyExists(Exception):
    pass


class AutoRefreshMixin:
    refreshed_fields = tuple()
    adapter = None
    EXPIRATION_TIME_MINUTES = 60

    def _get_new_expiration_registry(self):
        get_atr = super().__getattribute__
        expiration_registry = {item: None for item in get_atr("refreshed_fields")}
        self._expiration_registry = expiration_registry
        return expiration_registry

    def __getattribute__(self, item):
        get_atr = super().__getattribute__
        try:
            expiration_registry = get_atr("_expiration_registry")
        except AttributeError:
            expiration_registry = get_atr("_get_new_expiration_registry")()

        try:
            expiration_date = expiration_registry[item]
        except KeyError:
            return get_atr(item)

        is_expired = expiration_date is None or expiration_date <= datetime.datetime.utcnow()

        if is_expired:
            adapter = get_atr("adapter")
            refresh_func = getattr(adapter, item)
            value = refresh_func(self)  # TO CHECK: this will call getattribute again, can lead to loops

            expiration_registry[item] = datetime.datetime.utcnow() + datetime.timedelta(
                minutes=get_atr("EXPIRATION_TIME_MINUTES")
            )
            setattr(self, item, value)

        return get_atr(item)



from edclasses import OneToManyRelation


class Parent:
    _children_relation = OneToManyRelation.create(
        parent_class_name="Parent", child_class_name="Child"
    )

    def __init__(self, children=None):
        self.children = children or []

    def _children_setter(self, value):
        self._children_relation.set_for_parent(self, value)

    def _children_getter(self):
        return self._children_relation.get_for_parent(self)

    children = property(fget=_children_getter, fset=_children_setter)


class Child:
    _parent_relation = OneToManyRelation.create(
        parent_class_name="Parent", child_class_name="Child"
    )

    def __init__(self, parent=None):
        self.parent = parent

    def _parent_setter(self, value):
        self._parent_relation.set_for_child(self, value)

    def _parent_getter(self):
        return self._parent_relation.get_for_child(self)

    parent = property(fget=_parent_getter, fset=_parent_setter)


class TestOneToManyRelation:
    class TestChildSide:
        def test_removing_parent_removes_link(self):
            parent = Parent()
            child = Child(parent=parent)
            assert child.parent == parent
            assert parent.children == [child]

            child.parent = None

            assert child.parent is None
            assert parent.children == []

        def test_adding_parent_creates_link(self):
            parent = Parent()
            child = Child()

            child.parent = parent

            assert child.parent is parent
            assert parent.children == [child]

        def test_replacing_parent_replaces_link(self):
            old_parent = Parent()
            new_parent = Parent()
            child = Child(parent=old_parent)
            assert old_parent is not new_parent
            assert child.parent is old_parent
            assert old_parent.children == [child]

            child.parent = new_parent

            assert child.parent is new_parent
            assert new_parent.children == [child]
            assert old_parent.children == []

    class TestParentSide:
        def test_removing_children_removes_links(self):
            child1 = Child()
            child2 = Child()
            parent = Parent(children=[child1, child2])
            assert child1.parent is parent
            assert child2.parent is parent

            parent.children = []

            assert child1.parent is None
            assert child2.parent is None

        def test_adding_children_adds_links(self):
            child1 = Child()
            child2 = Child()
            parent = Parent(children=[])

            parent.children = [child1, child2]

            assert child1.parent is parent
            assert child2.parent is parent

        def test_stealing_children_replaces_links(self):
            old_parent = Parent()
            child1 = Child(parent=old_parent)
            child2 = Child(parent=old_parent)
            new_parent = Parent()
            assert old_parent.children == [child1, child2]
            new_parent.children = [child2]

            assert child1.parent is old_parent
            assert child2.parent is new_parent
            assert new_parent.children == [child2]
            assert old_parent.children == [child1]
